/*
 *  Copyright (C) 2012 Khaldoon Ghanem
 *  See the file license.txt for copying permission.
 */

%{
    #include <string.h>
    #include <stdarg.h>
    #include <stdio.h>
    #include <string.h>
    #define SHADER_DIR "./shader/"
    int yylex(void);
    void yyerror(char *);
    void myprintf(char** , const char *, int, ...);
    char *textFileRead(char *fn);
    #define YYSTYPE char*
    int linenum = 1;    
    extern int func_num;
    #ifndef FUNC_MAX
    	#define FUNC_MAX 64
    #endif
    extern char* func_input[];
    extern char* funcs[];    
%}

%token FLOAT_I
%token INTEGER_I
%token FLOAT
%token INTEGER
%token VAR
%token ABS
%token ARG
%token SIN
%token COS
%token TAN
%token EXP
%token LOG
%token INVALID

%left '+' '-'
%left '*' '/'
%left '^'
%nonassoc UMINUS

%%


program:
        program expr '\n'
        {
            int size = strlen($2)+1;
            func_input[func_num] = malloc(size);
            memcpy(func_input[func_num], $2, size);
        
            myprintf(&funcs[func_num],
                    "%s\nvec2 f(vec2 z){ return %s; }",
                    2, textFileRead(SHADER_DIR"complex.frag"), $2);
            func_num++;
            if(func_num==FUNC_MAX)
            {
                printf(
                        "Line %d: The maximum number of functions (%d)\
                          has been read. Stopping further parsing\n",
                        linenum, FUNC_MAX);
                YYACCEPT;
            }
            linenum++;
        }

        | program '\n'
        { 
            linenum++;
        }
        
        | program error '\n'	    
        { 
        	printf("Line %d: Syntax Error! Input line will be ignored.\n",
        			linenum++); 
        }

        |
        ;

expr:
    VAR                     { myprintf(&$$, "z", 0);}	
    | re                    { myprintf(&$$, "vec2(%s, 0.0)", 1, $1);}   
    | im                    { myprintf(&$$, "vec2(0.0,%s)", 1, $1);}
    | '-' expr              { myprintf(&$$, "(-1.0 * %s)", 1, $2);}
      %prec UMINUS
								
    | re VAR                { myprintf(&$$, "%s * z", 1, $1);}        	                
    | '(' expr ')'          { myprintf(&$$, "(%s)", 1,$2);}
    | expr '^' INTEGER      { myprintf(&$$, "cPow(%s, %s)", 2,$1, $3);}
    | expr '^' '-' INTEGER  { myprintf(&$$, "cPow(%s, -%s)", 2,$1, $4);}
      %prec UMINUS
        						    
    | expr '*' expr         { myprintf(&$$, "cMul(%s, %s)", 2, $1, $3);}
    | expr '/' expr         { myprintf(&$$, "cDiv(%s, %s)", 2, $1, $3);}        
    | expr '+' expr         { myprintf(&$$, "%s + %s", 2, $1, $3);}
    | expr '-' expr         { myprintf(&$$, "%s - %s", 2, $1, $3);}
    | SIN '(' expr ')'      { myprintf(&$$, "cSin(%s)", 1, $3);}
    | COS '(' expr ')'      { myprintf(&$$, "cCos(%s)", 1, $3);}
    | TAN '(' expr ')'      { myprintf(&$$, "cTan(%s)", 1, $3);}
    | EXP '(' expr ')'      { myprintf(&$$, "cExp(%s)", 1, $3);}
    | LOG '(' expr ')'      { myprintf(&$$, "cLog(%s)", 1, $3);}                                        
	;

re: FLOAT                   {$$ = $1;}
    | INTEGER               {myprintf(&$$, "%s.0", 1, $1);}
    | ABS '(' expr ')'      {myprintf(&$$, "cAbs(%s)", 1, $3);}
    | ARG '(' expr ')'      {myprintf(&$$, "cArg(%s)", 1, $3);}       		
	;
	
im: FLOAT_I                 {$$ = $1;}
    | INTEGER_I             {myprintf(&$$, "%s.0", 1, $1);}	
    ;	
%%

void myprintf(char** s, const char * format, int num, ...)
{
    va_list arglist;

    va_start( arglist, num );
    size_t size = vsnprintf(NULL, 0, format, arglist );
    va_end( arglist );
    *s = malloc(sizeof(char)*(size+1));

    va_start( arglist, num );
    vsprintf(*s, format, arglist);
    va_end( arglist );

    va_start( arglist, num );
    for ( int i = 0; i < num; i++ )
    free(va_arg ( arglist, char* ));
    va_end( arglist );
}

void yyerror(char *s)
{

}

