/*
 *  Copyright (C) 2012 Khaldoon Ghanem
 *  See the file license.txt for copying permission.
 */
 
vec4 HSVtoRGB(float h, float s, float v)
{
    //Convert between the HSV and RGB color model
    //Taken from http://www.cs.rit.edu/~ncs/color/t_convert.html and rewritten for GLSL
    int i;
    float f, p, q, t;
    vec4 RGB;

    if (s == 0.0)
    {
        // achromatic (grey)
        RGB.x = RGB.y = RGB.z = v;
        RGB.w = 1.0;
        return RGB;
    }

    h /= 60.0; // sector 0 to 5
    i = int(floor(h));
    f = h - float(i); // factorial part of h
    p = v * (1.0 - s);
    q = v * (1.0 - s * f);
    t = v * (1.0 - s * (1.0 - f));

   if(i==0)
    {
        RGB.x = v;
        RGB.y = t;
        RGB.z = p;
    }
    else if(i==1)
    {
        RGB.x = q;
        RGB.y = v;
        RGB.z = p;
    }
    else if(i==2)
    {
        RGB.x = p;
        RGB.y = v;
        RGB.z = t;
    }
    else if(i==3)
    {
        RGB.x = p;
        RGB.y = q;
        RGB.z = v;
    }
    else if(i==4)
    {
        RGB.x = t;
        RGB.y = p;
        RGB.z = v;
    }
    else if(i==5)
    {
        RGB.x = v;
        RGB.y = p;
        RGB.z = q;
    }
    else
    {
        RGB.x = 1.0;
        RGB.y = 1.0;
        RGB.z = 1.0;
    }
    RGB.w = 1.0;
    return RGB;
}

float PI = atan(1.0)*4.0;
float cAbs(vec2 z);
float cArg(vec2 z);

vec4 complexToColor(vec2 w)
{
    
    float r = cAbs(w);
    float brightness = fract(log2(r));
    
    //Compute color hue on color wheel
    float phi = cArg(w);
    if (phi < 0.0)
        phi += 2.0 * PI;
    float hue = degrees(phi);
    
    //Use full saturation
    return HSVtoRGB(hue, 1.0, brightness);
}
